/*
    Copyright 2016 Gleb Popov <6yearold@gmail.com>

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:
    1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
    3. Neither the name of the copyright holder nor the names of its contributors may
    be used to endorse or promote products derived from this software without specific
    prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
    INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
    OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
    ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "devdthread.h"

#include <QCoreApplication>
#include <QDebug>
#include <QHash>
#include <QRegularExpression>

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sys/socket.h>
#include <sys/un.h>

void DevdThread::parseLine(QString line)
{
    // !system=DEVFS subsystem=CDEV type=CREATE cdev=da0
    auto message = line.split(' ');
    QHash<QString, QString> data;
    Q_FOREACH(auto m, message)
    {
        auto mm = m.split('=');

        //skip things like "on", "at", etc.
        // +umass0 at bus=0 hubaddr=1 port=2 <...> intprotocol=0x50 on uhub0
        if(mm.count() == 1)
            continue;

        data.insert(mm[0], mm[1]);
    }

    QString devArg = data["cdev"].trimmed();

    if (QCoreApplication::arguments().contains("--debug-devd"))
        qDebug() << "devd message: " << data;


    // !system=GEOM subsystem=DEV type=CREATE cdev=da0
    // GEOM messages are always appear after corresponding DEVFS ones, so
    // we are looking for them only
    if(data["!system"] == "GEOM" &&
        data["subsystem"] == "DEV")
    {
        if(data["type"] == "CREATE")
            emit blockAdded(devArg);
        if(data["type"] == "DESTROY")
            emit blockRemoved(devArg);
        if(data["type"] == "MEDIACHANGE")
            emit blockChanged(devArg);

        return;
    }

    // !system=VFS subsystem=FS type=UNMOUNT
    // mount-point="/mnt" mount-dev="/dev/label/foo"
    if(data["!system"] == "VFS" &&
        data["subsystem"] == "FS")
    {
        // skip ZFS for now
        if(data["mount-type"] == "\"zfs\"")
            return;

        QString mountPoint = data["mount-point"].trimmed().chopped(1).remove(0,1);
        QString mountDev = data["mount-dev"].trimmed().chopped(1).remove(0,1);

        if(data["type"] == "UNMOUNT")
            emit fsUnmounted(mountPoint);
        if(data["type"] == "MOUNT")
            emit fsMounted(mountDev, mountPoint);
        return;
    }
}

void DevdThread::handle()
{
    struct sockaddr_un addr;
    int res;
    int s = socket(PF_LOCAL, SOCK_SEQPACKET, 0);

    addr.sun_family = PF_LOCAL;
    strcpy(addr.sun_path, "/var/run/devd.seqpacket.pipe");

    res = ::connect(s, (struct sockaddr*)&addr, sizeof(addr));
    if(res == -1)
    {
        qDebug() << "Unable to connect to /var/run/devd.seqpacket.pipe: " << strerror(errno);
        return;
    }

    while(true)
    {
        char buf[2048];
        int res = recv(s, buf, 2048, 0);
        if(res == -1)
        {
            qDebug() << "Unable to recv() from /var/run/devd.seqpacket.pipe: " << strerror(errno);
            shutdown(s, SHUT_RDWR);
            close(s);
            return;
        }
        if(res == 0)
        {
            qDebug() << "Connection to /var/run/devd.seqpacket.pipe was shut down";
            shutdown(s, SHUT_RDWR);
            close(s);
            return;
        }

        buf[res] = 0;
        parseLine(QString::fromLatin1(buf));
    }
}

void DevdThread::run()
{
    while(true)
    {
        handle();
        qDebug() << "Attempting to reconnect to /var/run/devd.seqpacket.pipe";
        QThread::sleep(1);
    }
}
